from .exceptions import InvalidFlag


class Flags:

    escape_ansi_code = '\x1b['
    escape_ansi_end = 'm'

    start_foreground = '38;5;'

    #### COLORS

    # Gray Scale
    BLACK = 'k', start_foreground + '16'
    GRAY = 'g', start_foreground + '243'
    WHITE = 'w', start_foreground + '15'

    # Red Scale
    DARKRED = '<r', start_foreground + '88'
    RED = 'r', start_foreground + '196'
    LIGHTRED = 'r>', start_foreground + '9'

    # Green Scale
    DARKGREEN = '<n', start_foreground + '22'
    GREEN = 'n', start_foreground + '28'
    LIGHTGREEN = 'n>', start_foreground + '65'

    # Yellow Scale
    DARKYELLOW = '<y', start_foreground + '58'
    YELLOW = 'y', start_foreground + '11'
    LIGHTYELLOW = 'y>', start_foreground + '3'

    # Blue Scale
    DARKBLUE = '<b', start_foreground + '17'
    BLUE = 'b', start_foreground + '20'
    LIGHTBLUE = 'b>', start_foreground + '4'

    # Magenta Scale
    DARKMAGENTA = '<m', start_foreground + '125'
    MAGENTA = 'm', start_foreground + '198'
    LIGHTMAGENTA = 'm>', start_foreground + '13'

    # Cyan Scale
    DARKCYAN = '<c', start_foreground + '30'
    CYAN = 'c', start_foreground + '39'
    LIGHTCYAN = 'c>', start_foreground + '51'

    # Orange Scale
    DARKORANGE = '<o', start_foreground + '130'
    ORANGE = 'o', start_foreground + '208'
    LIGHTORANGE = 'o>', start_foreground + '214'

    # Purple Scale
    DARKPURPLE = '<p', start_foreground + '54'
    PURPLE = 'p', start_foreground + '91'
    LIGHTPURPLE = 'p>', start_foreground + '98'

    #### FORMATS
    BOLD = 'B', '1'
    DIM = 'D', '2'
    ITALIC = 'I', '3'
    UNDERLINE = 'U', '4'
    HIGHLIGHT = 'H', '7'
    STRIKE = 'S', '9'

    #### END OF LINE
    reset = '0'

    @classmethod
    def get_end_of_line(cls):
        """ Defined method to get the 'reset' code """
        return cls.escape_ansi_code + cls.reset + cls.escape_ansi_end

    @classmethod
    def join_flags(cls, flags):
        """ Given a set of flags, returned the final ansi code to add to the text"""
        return "%s%s%s" % (cls.escape_ansi_code, ';'.join(flags), cls.escape_ansi_end)

    @classmethod
    def get_flags(cls):
        """
        returns a dictionary where the flag is the key and the attribute
        name is the value
        """
        return {y[0]: x for x, y in vars(cls).items() if x.isupper() and isinstance(y, tuple)}

    @classmethod
    def get_flag_values(cls, flags):
        """ returns a list of the escaped values for the flag labels """
        available_flags = cls.get_flags()
        flags_values = []

        # New in v1.3.0
        # flags can get a dark or a light intensity through the '<' and '>'
        # characters respectively. Dark colors have the form: <color, i.e. <b
        # light colors have the form color>, i.e. b>
        potential_flag = []
        flags = flags.replace(' ', '')  # remove white-spaces
        # We use inspect so we can check the following character also
        for f in range(len(flags)):
            flag = flags[f]

            if flag == '<':
                potential_flag.append(flag)
                continue
            elif flag == '>':
                potential_flag.append(flag)
                flag = ''.join(potential_flag)
                potential_flag.clear()
            else:
                try:
                    next_flag = flags[f + 1]
                except IndexError:
                    # the last one
                    next_flag = None

                if next_flag == '>':
                    potential_flag.append(flag)
                    continue
                else:
                    potential_flag.append(flag)
                    flag = ''.join(potential_flag)
                    potential_flag.clear()

            if flag not in available_flags:
                raise InvalidFlag(flag)
            else:
                if hasattr(cls, available_flags[flag]):
                    flags_values.append(getattr(cls, available_flags[flag])[1])
        return flags_values



import unittest
from printy.exceptions import InvalidFlag, InvalidInputType


class TestExceptions(unittest.TestCase):
    """ Test Case for exceptions """

    def setUp(self):
        self.invalid_flag = "l"
        self.invalid_flag_error = "'%s' is not a valid flag" % self.invalid_flag

        self.invalid_input_type = 'path'
        self.invalid_input_type_error = "'%s' is not a valid type" % self.invalid_input_type

    def test_invalid_flag_str(self):
        """ test that the exception InvalidFlag returns the expected text """
        error = InvalidFlag(self.invalid_flag)

        self.assertEqual(str(error), self.invalid_flag_error)

    def test_invalid_input_type_str(self):
        """ test that the exception InvalidInputType returns the expected text """
        error = InvalidInputType(self.invalid_input_type)

        self.assertEqual(str(error), self.invalid_input_type_error)



import platform

from .exceptions import InvalidInputType
from .flags import Flags

LINUX = 'Linux'
WINDOWS = 'Windows'
OSX = 'Darwin'

# For format() and format_input()
default_end = '\n'


class Printy:
    """
    Applies a format to the output of the print statement according
    to the flag (or flags).

    We can either set a global set of flags like >>> printy('Some text', 'rB')
    or set inline formats with the especial characters
    like >>> printy('[rB]Some@ [y]text@')
    """

    # For inline formatting we'll use special characters to catch the flags
    end_format_char = '@'
    open_flag_char = '['
    close_flag_char = ']'
    special_chars = [end_format_char, open_flag_char, close_flag_char]

    # Actions for inline formats
    START_FLAGS = 'start_flags'
    START_FORMAT = 'start_format'
    END_FORMAT = 'end_format'
    ESCAPE_CHAR = 'escape_char'

    def __init__(self):
        self.platform = platform.system()
        self.virtual_terminal_processing = self.set_windows_console_mode()

    def set_windows_console_mode(self):
        """
        For Windows os to work and get the escape sequences correctly,
        we'll need to enable the variable ENABLE_VIRTUAL_TERMINAL_PROCESSING

        @Thanks to Mihir Singh (mihirs16) for this big improvement
        """
        # In case there is some error while setting it up, we returns False to
        # indicate that windows will not print the escape sequences correctly,
        # so we can print out the cleaned text
        if self.platform == WINDOWS:
            try:
                from ctypes import windll
                k = windll.kernel32
                k.SetConsoleMode(k.GetStdHandle(-11), 7)
                return True
            except ImportError:
                return False
        return False

    @classmethod
    def _define_char(cls, prev: str, current: str) -> bool:
        """
        Helper method that'll tell us if a character has to be treated as a
        special one or it is part of the text that 's intended to be printed
        out.

        Takes the previous character and the current character, in case it is
        one of the special characters defined in the class and is prepended by
        a '\' means that it has not to be treated as a special one

        Returns True if it's a special character
        """
        if current in cls.special_chars:
            if prev != '\\':
                return True
        return False

    @classmethod
    def _check_special_char_position(cls, last_special: str, special: str) -> str:
        """
        Returns an action to execute if the character is well placed. It should
        only be applied over special characters.

        If it's not well placed, the character will be included in the text
        """
        if special == cls.open_flag_char:
            # In this case the 'last_special' must always be
            # the 'end_format_char' or None if it's the first appearing
            if last_special not in [cls.end_format_char, None]:
                return cls.ESCAPE_CHAR
            else:
                return cls.START_FLAGS
        elif special == cls.close_flag_char:
            # In this case the 'last_special' must always be the 'open_flag_char'
            if last_special != cls.open_flag_char:
                return cls.ESCAPE_CHAR
            else:
                return cls.START_FORMAT
        elif special == cls.end_format_char:
            # In this case the 'last_special' must always be the 'close_flag_char'
            # Or None if the text does not include any other formatting character
            if last_special in [cls.open_flag_char, cls.end_format_char, None]:
                return cls.ESCAPE_CHAR
            else:
                return cls.END_FORMAT

    @classmethod
    def _replace_escaped(cls, text: str) -> str:
        """ Replaces escaped special characters for the character itself """
        for special_char in cls.special_chars:
            text = text.replace('\\' + special_char, special_char)
        return text

    @classmethod
    def _get_inline_format_as_tuple(cls, text: str) -> list:
        """
        In case some inline formats have been applied we need to get a list of
        tuples indicating the formats to be applied via flags and the text
        where the format should be applied, for instance, if the text is:
        "[rB]Some@ Te[H]xt@"
        We'll get the list [('Some', 'rB), (' Te', None), ('xt', 'H')]
        """
        prev = ''  # Stores the last character in the loop
        last_special_char = None
        list_of_formats = []  # Final list to be returned

        # While looping the text, we'll get the formats to be applied to
        # certain section of that text
        section_text = []
        section_flags = []
        current_action = cls.START_FORMAT

        # Will tell us when we're at the last character of the loop
        counter = 0
        # Will tell us if we need to reset the sections variables
        close_section = False
        for char in text:
            is_special = cls._define_char(prev, char)
            if is_special:
                action = cls._check_special_char_position(last_special_char,
                                                          char)

                if action == cls.ESCAPE_CHAR:
                    # Add the character to the text
                    if current_action == cls.START_FLAGS:
                        section_flags.append(char)
                    elif current_action == cls.START_FORMAT:
                        section_text.append(char)
                else:
                    # Here we know that the special character is well placed
                    # and has a special meaning
                    current_action = action

                    if current_action == cls.END_FORMAT:
                        current_action = cls.START_FORMAT
                        close_section = True
                    if last_special_char in [cls.end_format_char, None]:
                        # Here we'll catch 'open_flag_char's
                        current_action = cls.START_FLAGS
                        close_section = True
                    last_special_char = char
            else:
                if current_action == cls.START_FLAGS:
                    section_flags.append(char)
                elif current_action == cls.START_FORMAT:
                    section_text.append(char)
            prev = char
            counter += 1

            if counter == len(text) or close_section:
                # Reset the 'section_*' lists and add them (joined)
                # to the final list
                list_of_formats.append((
                    ''.join(section_text),
                    ''.join(section_flags) if len(section_flags) > 0 else None
                ))
                section_text = []
                section_flags = []
                close_section = False

        return list_of_formats

    @classmethod
    def _get_cleaned_text(cls, text: str) -> str:
        """ Returns the cleaned value, with no formats """
        tuple_text = cls._get_inline_format_as_tuple(text)
        return cls._replace_escaped(''.join(x[0] for x in tuple_text))

    @classmethod
    def _escape_special_chars(cls, value):
        """
        Escape all the special characters in the value (or the string
        representation of the value if an object is passed)
        """
        _str = str(value)
        for char in cls.special_chars:
            _str = _str.replace(char, '\\' + char)
        return _str

    @classmethod
    def _pretty_print_object(cls, obj, indentation: int, level: int = 1) -> str:
        """
        Recursive function to pretty print objects, with some
        indentations if needed.
        """

        def _nested(nested_obj, nested_level: int = level) -> str:
            indentation_values = " " * indentation * nested_level
            indentation_braces = " " * indentation * (nested_level - 1)
            if isinstance(nested_obj, dict):
                return "{\n%(body)s%(indent_braces)s}" % {
                    "body": "".join("%(indent_values)s[n>]\'%(key)s\'@: %(value)s[<oB],@\n" % {
                        "key": str(key),
                        "value": _nested(value, nested_level + 1),
                        "indent_values": indentation_values
                    } for key, value in nested_obj.items()),
                    "indent_braces": indentation_braces
                }
            elif isinstance(nested_obj, list):
                return "\[\n%(indent_values)s%(body)s\n%(indent_braces)s\]" % {
                    "body": "[<oB],@ ".join("%(value)s" % {
                        "value": _nested(value, nested_level + 1),
                    } for value in nested_obj),
                    "indent_braces": indentation_braces,
                    "indent_values": indentation_values
                }
            elif isinstance(nested_obj, tuple):
                return "(\n%(indent_values)s%(body)s\n%(indent_braces)s)" % {
                    "body": "[<oB],@ ".join("%(value)s" % {
                        "value": _nested(value, nested_level + 1),
                    } for value in nested_obj),
                    "indent_braces": indentation_braces,
                    "indent_values": indentation_values
                }
            elif isinstance(nested_obj, set):
                return "{\n%(indent_values)s%(body)s\n%(indent_braces)s}" % {
                    "body": "[<oB],@ ".join("%(value)s" % {
                        "value": _nested(value, nested_level + 1),
                    } for value in nested_obj),
                    "indent_braces": indentation_braces,
                    "indent_values": indentation_values
                }
            else:
                if isinstance(nested_obj, str):
                    return "[c>]\'" + cls._escape_special_chars(nested_obj) + "\'@"
                elif isinstance(nested_obj, bool) or nested_obj is None:
                    return "[<o]" + cls._escape_special_chars(nested_obj) + "@"
                elif isinstance(nested_obj, (int, float)):
                    return "[c]" + str(nested_obj) + "@"
                else:
                    return str(nested_obj)
        return _nested(obj)

    @classmethod
    def _repr_value(cls, value, pretty: bool = True, indentation: int = 4) -> str:
        """
        In case a dictionary or a list or a set is passed as a value, we need
        to get the representation of it, and, escape all the '[', ']' and '@'
        characters that we find
        """
        if isinstance(value, (dict, list, tuple, set)):
            if pretty:
                return cls._pretty_print_object(value, indentation)
            else:
                return cls._escape_special_chars(value)
        elif isinstance(value, bool) or value is None:
            return "[<o]" + str(value) + "@"
        elif isinstance(value, (int, float)):
            return "[c]" + str(value) + "@"
        else:
            return str(value)

    def get_formatted_text(self, value: str, flags: str = '', predefined: str = '',
                           pretty: bool = True, indentation: int = 4, **kwargs) -> str:
        """
        Applies the format specified by the 'flags' to the 'value'.

        If 'flag's is passed, 'predefined' will be omitted.
        """
        # In case an object is passed instead of a string
        value = self._repr_value(value, pretty, indentation)

        if self.platform == WINDOWS and not self.virtual_terminal_processing:
            text = self._get_cleaned_text(value)
        else:
            if flags:
                flags_values = Flags.get_flag_values(flags)
                value = self._get_cleaned_text(value)
                text = "%s%s%s" % (
                    Flags.join_flags(flags_values),
                    value,
                    Flags.get_end_of_line()
                )
            else:
                tuple_text = self._get_inline_format_as_tuple(value)
                text = ''
                for section in tuple_text:
                    section_text = self._replace_escaped(section[0])
                    section_flags = section[1] or predefined
                    if section_flags:
                        flags_values = Flags.get_flag_values(section_flags)
                        text += "%s%s%s" % (
                            Flags.join_flags(flags_values),
                            section_text,
                            Flags.get_end_of_line()
                        )
                    else:
                        text += section_text
        return text

    @staticmethod
    def read_file(file: str) -> str:
        """ Given a file path, we read it and print it out """
        file = str(file)
        with open(file) as f:
            text = f.read()
        return text

    def format(self, value='', flags='', predefined='', file='',
               end=default_end, pretty=True, indentation=4):
        """ Prints out the value """
        value = self.read_file(file) if file else value
        print(self.get_formatted_text(
            value, flags, predefined, pretty, indentation), end=end
        )

    ##### ============= INPUTY ======================

    # Types (str is the default)
    BOOL = 'bool'
    INT = 'int'
    FLOAT = 'float'
    STR = 'str'
    types = [BOOL, INT, FLOAT, STR]

    def check_boolean(self, value: str, options: dict, condition: str) -> tuple:
        """
        Validates the value when the type must be a boolean, returns a boolean
        specifying whether it is a valid value, and if it is, returns the final
        value (after conversions if necessary)
        """

        true_value, false_value = options["1"], options["2"]
        error_msg = "[o]%s@ is not a valid value, enter %s or %s" % (
            value,
            true_value,
            false_value,
        )
        if condition == 'i':
            true_value = true_value.lower()
            false_value = false_value.lower()
            value = value.lower()

        if value == true_value:
            return True, True
        elif value == false_value:
            return False, True
        else:
            self.format(error_msg)
            return False, False

    def _check_number(self, number_class, value: str = '', condition: str = '',
                      max_digits: int = None, max_decimals: int = None) -> tuple:
        """ Validates the value when it is a number """

        # the only options allowed are '+' and '-'
        error_msg = (
            "\t[r>]Invalid Value:@ [o]%s@ is not a valid number" % value
        )
        # Let's try to convert it to the number type
        valid_value = False
        try:
            value = number_class(value)
        except (ValueError, TypeError):
            if number_class == int:
                error_msg += (
                    ".\n\tPlease enter a [b>]rounded@ number, please check you are "
                    "\n\tnot adding some [p>]decimal digits@"
                )
        else:
            if condition:
                if condition == '+' and value >= 0:
                    valid_value = True
                elif condition == '-' and value < 0:
                    valid_value = True
                else:
                    error_msg += ', \n\tMake sure it is a [y]%s@ number\n' % (
                        'positive' if condition == '+' else 'negative'
                    )
                    valid_value = False
            else:
                valid_value = True

            # Let's remove the negative sign if any
            str_value = str(value).replace('-', '')
            split_number = str_value.split('.')
            # Check max digits
            if max_digits is not None:
                if len(split_number[0]) > max_digits:
                    error_msg += '. \n\tThe number must have [o]%d@ digits max' % max_digits
                    valid_value = False
            if max_decimals is not None and len(split_number) > 1:
                if len(split_number[1]) > max_decimals:
                    error_msg += '. \n\tThe number must have [o]%d@ decimals max' % max_decimals
                    valid_value = False
        # Throw error
        if not valid_value:
            self.format(error_msg)

        return value, valid_value

    def check_integer(self, value: str, condition: str = '', max_digits: int = None) -> tuple:
        """
        Validates the value when the type must be an integer, returns a boolean
        specifying whether it is a valid value, and if it is, returns the final
        value (after conversions if necessary)
        """

        return self._check_number(
            int,
            value=value,
            condition=condition,
            max_digits=max_digits
        )

    def check_float(self, value: str, condition: str = '',
                    max_digits: int = None, max_decimals: int = None) -> tuple:
        """
        Validates the value when the type must be an float, similar to integer check,
        but now it can have decimal digits, returns a boolean specifying whether it
        is a valid value, and if it is, returns the final
        value (after conversions if necessary)
        """

        return self._check_number(
            float,
            value=value,
            condition=condition,
            max_digits=max_digits,
            max_decimals=max_decimals
        )

    def check_string(self, value: str, options: dict = None, condition: str = ''):
        """
        if options were passed, then it validates that the value belongs to
        those options
        """
        error_msg = "[o]%s@ is not a valid value" % value
        if options:
            if condition == 'i':
                # then we need to create a dictionary where the keys are the
                # lowercase of the options' values, and the values are the
                # options' keys
                new_options = {}
                for key, val in options.items():
                    new_options[val.lower()] = key
                value = value.lower()

                if value in new_options.keys():
                    return options[new_options[value]], True
                elif value in options.keys():
                    return options[value], True
                else:
                    self.format(error_msg)
                    return value, False
            else:
                if value in options.values():
                    return value, True
                elif value in options.keys():
                    return options[value], True
                else:
                    self.format(error_msg)
                    return value, False
        else:
            return value, True

    @classmethod
    def _render_options(cls, options: dict, input_type: str, default: str,
                        render_options: bool) -> str:
        """
        Returns a string to present to the user specifying the available
        options. 'options' must be normalized.
        """
        render = ""
        if render_options:
            if options and input_type in [cls.STR, cls.BOOL]:
                if input_type == cls.BOOL:
                    if len(options) >= 2:
                        render += " (%s/%s)" % (options['1'], options['2'])
                    if default:
                        # We escape the '[' and ']' so they can be formatted
                        render += " \[%s\]" % default
                else:
                    render += " default: %s\n" % default if default else '\n'
                    for item, value in options.items():
                        render += "  %s) %s\n" % (item, value)

        return render

    @classmethod
    def _normalize_options(cls, options: list, input_type: str):
        """
        Takes the list passed as options and returns a dictionary enumerating
        all the options, i.e. if we pass ['option_1', 'options_2'], then we
        return {'1': 'option_1', '2': 'option_2'}
        """
        normalized_options = {}
        if options is not None:
            if input_type == cls.BOOL:
                if len(options) < 2:
                    options = ['True', 'False']
                else:
                    options = options[:2]
            else:
                if len(options) < 2:
                    raise ValueError("'options' must contain at least two items")
        else:
            options = ['True', 'False'] if input_type == cls.BOOL else []

        for option in range(len(options)):
            normalized_options[str(option + 1)] = str(options[option])

        return normalized_options

    @staticmethod
    def _to_int(value):
        """
        Helper function to convert a value into an integer, useful
        to validate some parameters
        """
        if value is not None:
            try:
                value = int(value)
            except (ValueError, TypeError):
                raise

        return value

    def format_input(self, *args, **kwargs):
        """
        Colorize the text prompted by input() and applies some validation.

        Also, it takes an additional parameter 'type', to tell the prompt not
        to accept a format other than the specified. As every input is converted
        to strings, a string that can be converted to the specified type is
        allowed. For example, if type=int, then the user would be forced to
        enter a number or a string that can be converted into an integer
        """

        # If passed, we'll force the user to write a value with the specific
        # input_types' format.
        input_type = kwargs.get('type', self.STR)
        if input_type not in self.types:
            raise InvalidInputType(input_type)

        # A list containing the value valid values
        options = kwargs.get('options', None)
        # render_options defines whether the options should be rendered or not
        render_options = kwargs.get('render_options', True)
        # when no value is entered, the default will be added
        default = str(kwargs.get('default', ''))
        # Tells us the conditions according to the type, for numbers
        # we can say '+' or '-' whether we want only positives values or negative
        # values, for strings and boolean we can specify 'i' if we want to allow
        # user to type cas insensitive values
        condition = str(kwargs.get('condition', ''))
        # Defines the max number of digits that a number (int or float) can have
        max_digits = self._to_int(kwargs.get('max_digits', None))
        # Defines the max number of decipal numbers that the value can have,
        # only valid for 'float' type
        max_decimals = self._to_int(kwargs.get('max_decimals', None))

        # Include the value for the get_formatted_text function
        text, flags = '', ''
        if len(args) == 0:
            args = [text, flags]
        elif len(args) == 1:
            text = args[0]
        else:
            text, flags = args[0], args[1]

        # Normalize the options
        options = self._normalize_options(options, input_type)
        # Will tell us whether the user sent a value value according to the
        # specified type or not
        valid_value = False
        result = None
        while not valid_value:
            render_options = self._render_options(options, input_type, default, render_options)
            # Prints out the message if any was passed
            result = str(
                input(
                    self.get_formatted_text(*args, **kwargs)
                    + self.get_formatted_text(render_options, flags)
                )
            )

            if result == '':
                result = default

            if input_type == self.BOOL:
                result, valid_value = self.check_boolean(result, options, condition)

            elif input_type == self.INT:
                result, valid_value = self.check_integer(result, condition, max_digits)

            elif input_type == self.FLOAT:
                result, valid_value = self.check_float(result, condition, max_digits, max_decimals)

            else:
                result, valid_value = self.check_string(result, options, condition)

        return result



import unittest
from unittest import mock
from printy.exceptions import InvalidFlag, InvalidInputType
from printy.core import Printy, WINDOWS
from printy.flags import Flags


class TestGlobalFlagsPrinty(unittest.TestCase):
    """ Test case for formatting with a global set of flags specified """

    def setUp(self):
        self.sample_text = "Some Text To Print Out"
        self.printy = Printy()
        self.raw_text = self.printy.get_formatted_text

    def test_empty_value(self):
        """ Tests passing an empty value print's nothing"""
        text = ''
        result = self.raw_text(text)

        self.assertEqual(result, text)

    def test_empty_value_with_flags(self):
        """
        Tests that passing and empty value with some flags returns the
        escape ansi characters
        """
        text = ''
        flags = 'rBH'
        result = self.raw_text(text, flags)
        expected_result = "%s%s" % (
                Flags.join_flags(Flags.get_flag_values(flags)),
                Flags.get_end_of_line()
        )

        self.assertEqual(result, expected_result)

    def test_single_invalid_flag(self):
        """
        Tests that passing an invalid flag (only one)
        raises and exception
        """
        invalid_flag = 'P'
        with self.assertRaises(InvalidFlag):
            self.printy.format(self.sample_text, invalid_flag)

    def test_multiple_invalid_flag(self):
        """
        Tests that passing multiple invalid flags raises an
        exception with the first invalid flag found
        """
        # P and G are invalid, should raise InvalidFlag
        # with 'P' as invalid flag
        flags = 'yBPGr'
        with self.assertRaises(InvalidFlag) as e:
            self.printy.format(self.sample_text, flags)
        self.assertEqual(e.exception.flag, 'P')

    def test_high_intensity_flag_color(self):
        """
        Checks the correct format is returned for a high
        intensity (>) flag color
        """
        flag = 'p>'
        text = 'Hello'
        expected_text = '\x1b[38;5;98mHello\x1b[0m'

        self.assertEqual(self.raw_text(text, flag), expected_text)

    def test_low_intensity_flag_color(self):
        """
        Checks the correct format is returned for a low
        intensity (<) flag color
        """
        flag = '<p'
        text = 'Hello'
        expected_text = '\x1b[38;5;54mHello\x1b[0m'

        self.assertEqual(self.raw_text(text, flag), expected_text)

    def tests_always_closing_format(self):
        """
        Tests that the returned text is always ended with the closing tag
        """
        result = self.raw_text(self.sample_text, 'r')
        closing_tag = result[-4:]
        self.assertEqual(len(closing_tag), 4)
        self.assertEqual(closing_tag, Flags.get_end_of_line())

    def test_no_flag_parameter_passed(self):
        """
        Tests that passing no flag parameter will return a default value
        """
        result = self.raw_text(self.sample_text)
        self.assertEqual(result, self.sample_text)

    def test_empty_flag(self):
        """
        Test that passing and empty string as a flag still returns the
        default value
        """
        result = self.raw_text(self.sample_text, '')
        self.assertEqual(result, self.sample_text)

    def test_flags_with_spaces_in_between(self):
        """
        Tests that passing a set of flags with some spaces in between
        (like 'yB H U') still applies the desired formats
        """
        desired_flags = 'yBH'
        flags_with_one_space = 'yB H'
        flags_with_multiple_spaces = 'y B H'
        result_one = self.raw_text(self.sample_text, desired_flags)
        result_two = self.raw_text(self.sample_text, flags_with_one_space)
        result_three = self.raw_text(self.sample_text, flags_with_multiple_spaces)

        self.assertTrue(result_one == result_two == result_three)

    @mock.patch('printy.core.Printy.set_windows_console_mode', return_value=True)
    def test_virtual_terminal_processing_on_windows(self, mock_console_mode):
        """
        Tests that if platform is windows, then then returns True
        """
        self.printy.platform = WINDOWS
        virtual_terminal_processing = mock_console_mode()

        self.assertTrue(virtual_terminal_processing)

    def test_return_cleaned_value_if_windows_is_not_properly_configured(self):
        """
        Tests that if printy virtual_console_mode is false, then it returns the
        cleaned_text
        """
        flags = 'rBH'
        # Changes platform to Windows
        self.printy.platform = WINDOWS
        self.printy.virtual_terminal_processing = False

        result_one = self.raw_text(self.sample_text, flags)

        self.assertEqual(result_one, self.sample_text)


class TestInlineFlagsPrinty(unittest.TestCase):
    """ Test case for inline formatting """

    def setUp(self):
        self.printy = Printy()
        self.raw_text = self.printy.get_formatted_text

    def test_inline_format_with_global_flags(self):
        """
        Tests that passing a text with inline formatting and also a global
        set of flags takes this last one as the format to be applied
        """
        inline_formatted = "[y]Hey you@"
        no_format = 'Hey you'
        global_flags = 'rB'
        result_one = self.raw_text(inline_formatted, global_flags)
        result_two = self.raw_text(no_format, global_flags)

        self.assertEqual(result_one, result_two)

    def test_inline_format_without_ending_format_character(self):
        """
        Tests that passing an inline formatted text without the ending
        formatting character still returns the formatted text
        """
        result_one = self.raw_text('[y]Hey you')
        result_two = self.raw_text('[y]Hey you@')

        self.assertEqual(result_one, result_two)

    def test_escape_special_characters(self):
        """ Tests that escaping special characters prints them out """
        inline_text_one = '[y]myemail\@mydomain.com@'
        global_text_one = 'myemail@mydomain.com', 'y'

        inline_text_two = '[bH]Some text \@@'
        global_text_two = 'Some text @', 'bH'

        inline_result_one = self.raw_text(inline_text_one)
        global_result_one = self.raw_text(global_text_one[0], global_text_one[1])

        inline_result_two = self.raw_text(inline_text_two)
        global_result_two = self.raw_text(global_text_two[0], global_text_two[1])

        self.assertEqual(inline_result_one, global_result_one)
        self.assertEqual(inline_result_two, global_result_two)

    def test_multiple_sections(self):
        """ Test that formats are applied correctly to each section """

        section_one = "Some"
        section_two = ' '
        section_three = 'text'
        global_format_one = self.raw_text(section_one, 'rB')
        global_format_two = self.raw_text(section_two)
        global_format_three = self.raw_text(section_three, 'y')
        joined_global_format = global_format_one + global_format_two + global_format_three

        inline_text = '[rB]Some@ [y]text@'
        inline_format = self.raw_text(inline_text)

        self.assertEqual(inline_format, joined_global_format)

    def test_read_file(self):
        """ Test retrieving the text from a file """
        text_in_file = 'printy'
        file_name = 'printy_file'
        with mock.patch('builtins.open', mock.mock_open(read_data=text_in_file)) as m:
            result = self.printy.read_file(file_name)

        m.assert_called_once_with(file_name)
        self.assertEqual(result, text_in_file)

    def test_escape_special_chars_method(self):
        """
        Test escaping especial characters correctly, this method is used when
        an object other than a string is passed
        """
        text_to_escape = '[some text @ ]'
        expected_value = '\[some text \@ \]'
        escaped_text = Printy._escape_special_chars(text_to_escape)

        self.assertEqual(expected_value, escaped_text)

    def test_pretty_print_dicts(self):
        """ Test pretty printing dictionaries """
        dict_to_print = {'name': 'John Doe', 'age': 34}
        expected_result = '{\n    [n>]\'name\'@: [c>]\'John Doe\'@[<oB],@\n    [n>]\'age\'@: [c]34@[<oB],@\n}'
        pretty_dict = Printy._repr_value(dict_to_print)

        self.assertEqual(expected_result, pretty_dict)

    def test_pretty_print_lists(self):
        """ Test pretty printing lists """
        list_to_print = [1, 2, 'hello']
        expected_result = '\[\n    [c]1@[<oB],@ [c]2@[<oB],@ [c>]\'hello\'@\n\]'
        pretty_list = Printy._repr_value(list_to_print)

        self.assertEqual(expected_result, pretty_list)

    def test_pretty_printy_tuples(self):
        """ Test pretty printing tuples """
        tuple_to_print = (1, 2, 'hello')
        expected_result = '(\n    [c]1@[<oB],@ [c]2@[<oB],@ [c>]\'hello\'@\n)'
        pretty_tuple = Printy._repr_value(tuple_to_print)

        self.assertEqual(expected_result, pretty_tuple)

    def test_pretty_printy_sets(self):
        """ Test pretty printing sets """
        set_to_print = {1, 2, 'hello'}
        expected_result = '{\n    [c]1@[<oB],@ [c]2@[<oB],@ [c>]\'hello\'@\n}'
        pretty_set = Printy._repr_value(set_to_print)

        self.assertEqual(expected_result, pretty_set)

    def test_pretty_printy_dict_pretty_false(self):
        """ Tests pretty printing a dict when 'pretty' parameter is set to False """
        dict_to_print = {'name': 'John Doe', 'age': 34}
        expected_result = '{\'name\': \'John Doe\', \'age\': 34}'
        not_pretty_dict = Printy._repr_value(dict_to_print, pretty=False)

        self.assertEqual(expected_result, not_pretty_dict)

    def test_pretty_printy_list_pretty_false(self):
        """ Tests pretty printing a list when 'pretty' parameter is set to False """
        list_to_print = [1, 2, 'hello']
        expected_result = '\[1, 2, \'hello\'\]'
        not_pretty_list = Printy._repr_value(list_to_print, pretty=False)

        self.assertEqual(expected_result, not_pretty_list)

    def test_pretty_printy_tuple_pretty_false(self):
        """ Tests pretty printing a tuple when 'pretty' parameter is set to False """
        tuple_to_print = (1, 2, 'hello')
        expected_result = '(1, 2, \'hello\')'
        not_pretty_tuple = Printy._repr_value(tuple_to_print, pretty=False)

        self.assertEqual(expected_result, not_pretty_tuple)

    def test_pretty_printy_set_pretty_false(self):
        """ Tests pretty printing a set when 'pretty' parameter is set to False """
        set_to_print = {1, 2, 'hello'}
        expected_result = '{1, 2, \'hello\'}'
        not_pretty_set = Printy._repr_value(set_to_print, pretty=False)

        self.assertEqual(expected_result, not_pretty_set)

    def test_pretty_print_str_method_of_objects(self):
        """ Test printing the str method of an object, both not defined and defined """
        builtin_obj = int
        expected_builtin_result = '<class \'int\'>'
        pretty_builtin = Printy._repr_value(builtin_obj)

        class Person:
            def __str__(self):
                return '[c]I am a person@'
        custom_str = Person()
        # Notice how it should not return the escaped character
        expected_custom_result = '[c]I am a person@'
        pretty_custom = Printy._repr_value(custom_str)

        self.assertEqual(expected_builtin_result, pretty_builtin)
        self.assertEqual(expected_custom_result, pretty_custom)

    def test_pretty_object_in_dictionary(self):
        """
        Test pretty printing an str method of an object inside a dictionary
        or any iterable, it should give it a light magenta color
        """
        dict_to_print = {'class': int}
        expected_result = '{\n    [n>]\'class\'@: <class \'int\'>[<oB],@\n}'
        pretty_dict = Printy._repr_value(dict_to_print)

        self.assertEqual(expected_result, pretty_dict)

    def test_pretty_custom_str_method_in_dictionary(self):
        class CustomStrMethod:
            def __str__(self):
                return '[rBU]Red Bold Underlined@ and [y]Yellow@'

        dict_to_print = {'str': CustomStrMethod()}
        expected_result = '{\n    [n>]\'str\'@: [rBU]Red Bold Underlined@ and [y]Yellow@[<oB],@\n}'
        pretty_dict = Printy._repr_value(dict_to_print)

        self.assertEqual(expected_result, pretty_dict)

    def test_print_number(self):
        integer_to_print = 123
        float_to_print = 123.45
        expected_result_integer = '[c]123@'
        expected_result_float = '[c]123.45@'

        result_integer = Printy._repr_value(integer_to_print)
        result_float = Printy._repr_value(float_to_print)

        self.assertEqual(expected_result_integer, result_integer)
        self.assertEqual(expected_result_float, result_float)

    def test_print_boolean(self):
        expected_false = '[<o]False@'
        expected_true = '[<o]True@'

        result_false = Printy._repr_value(False)
        result_true = Printy._repr_value(True)

        self.assertEqual(expected_false, result_false)
        self.assertEqual(expected_true, result_true)

    def test_print_none(self):
        expected_none = '[<o]None@'
        result_none = Printy._repr_value(None)

        self.assertEqual(expected_none, result_none)


class TestInputy(unittest.TestCase):
    """
    Test case for inputy functionality

    Here, it is not necessary to test whether the prompted message has the
    correct format because it uses the methods already tested in the Printy
    test cases
    """

    def setUp(self):
        self.inputy = Printy()

    str_valid_test = "Valid String"
    int_valid_test = 23
    float_valid_test = 45.6
    bool_valid_test = False

    def test_normalize_options_not_enough_options_raises_value_error(self):
        """
        Tests that passing a list with less than 2 items as 'options' raises
        a Value Error when input_type is different than 'bool'
        """
        with self.assertRaises(ValueError):
            self.inputy._normalize_options([], self.inputy.STR)

    def test_normalize_options_bool_more_than_two_items_get_the_first_two(self):
        """
        Tests that passing more than 2 items as options for input_type 'bool'
        returns only the first two
        """
        options = ['yes', 'no', 'Yep', 'Nope']
        normalized_options = self.inputy._normalize_options(options, self.inputy.BOOL)
        expected_options = {'1': 'yes', '2': 'no'}

        self.assertEqual(normalized_options, expected_options)

    def test_normalize_options_bool_less_than_two_items_returns_tru_false(self):
        """
        Tests that passing more than 2 items as options for input_type 'bool'
        returns only the first two
        """
        options = ['yes']
        normalized_options = self.inputy._normalize_options(options, self.inputy.BOOL)
        expected_options = {'1': 'True', '2': 'False'}

        self.assertEqual(normalized_options, expected_options)

    def test_check_boolean_case_insensitive(self):
        """
        Test that passing a different case for one of the options still
        returns the value
        """
        options = {'1': 'y', '2': 'n'}
        value = 'Y'
        result, valid = self.inputy.check_boolean(value, options, 'i')

        self.assertEqual(result, True)
        self.assertEqual(valid, True)

    def test_check_boolean_case_sensitive(self):
        """
        Test that passing a different case for one of the options returns
        an invalid value
        """
        options = {'1': 'y', '2': 'n'}
        value = 'Y'
        result, valid = self.inputy.check_boolean(value, options, '')

        self.assertEqual(result, False)
        self.assertEqual(valid, False)

    def test_check_integer_no_condition(self):
        """
        Tests that passing no condition to check an integer returns the value
        """
        positive_value = 5
        negative_value = -5
        condition = ''
        result_positive, valid_positive = self.inputy.check_integer(str(positive_value), condition)
        result_negative, valid_negative = self.inputy.check_integer(str(negative_value), condition)

        self.assertTrue(valid_positive)
        self.assertTrue(valid_negative)
        self.assertEqual(result_positive, 5)
        self.assertEqual(result_negative, -5)

    def test_check_integer_condition_only_positive(self):
        """
        Test that passing a condition '-' to the check_integer function will
        return a valid value only when the value is negative
        """
        positive_value = 5
        negative_value = -5
        condition = '-'
        result_positive, valid_positive = self.inputy.check_integer(str(positive_value), condition)
        result_negative, valid_negative = self.inputy.check_integer(str(negative_value), condition)

        self.assertFalse(valid_positive)
        self.assertTrue(valid_negative)

        self.assertEqual(result_positive, 5)
        self.assertEqual(result_negative, -5)

    def test_check_integer_condition_only_negative(self):
        """
        Test that passing a condition '-' to the check_integer function will
        return a valid value only when the value is negative
        """
        positive_value = 5
        negative_value = -5
        condition = '+'
        result_positive, valid_positive = self.inputy.check_integer(str(positive_value), condition)
        result_negative, valid_negative = self.inputy.check_integer(str(negative_value), condition)

        self.assertTrue(valid_positive)
        self.assertFalse(valid_negative)

        self.assertEqual(result_positive, 5)
        self.assertEqual(result_negative, -5)

    def test_check_float_no_condition(self):
        """
        Tests that passing no condition to check a float returns the value
        """
        positive_value = 5.0
        negative_value = -5.0
        condition = ''
        result_positive, valid_positive = self.inputy.check_float(str(positive_value), condition)
        result_negative, valid_negative = self.inputy.check_float(str(negative_value), condition)

        self.assertTrue(valid_positive)
        self.assertTrue(valid_negative)
        self.assertEqual(result_positive, 5.0)
        self.assertEqual(result_negative, -5.0)

    def test_check_float_condition_only_positive(self):
        """
        Test that passing a condition '-' to the check_float function will
        return a valid value only when the value is negative
        """
        positive_value = 5.0
        negative_value = -5.0
        condition = '-'
        result_positive, valid_positive = self.inputy.check_float(str(positive_value), condition)
        result_negative, valid_negative = self.inputy.check_float(str(negative_value), condition)

        self.assertFalse(valid_positive)
        self.assertTrue(valid_negative)

        self.assertEqual(result_positive, 5.0)
        self.assertEqual(result_negative, -5.0)

    def test_check_float_condition_only_negative(self):
        """
        Test that passing a condition '-' to the check_float function will
        return a valid value only when the value is negative
        """
        positive_value = 5.0
        negative_value = -5.0
        condition = '+'
        result_positive, valid_positive = self.inputy.check_float(str(positive_value), condition)
        result_negative, valid_negative = self.inputy.check_float(str(negative_value), condition)

        self.assertTrue(valid_positive)
        self.assertFalse(valid_negative)

        self.assertEqual(result_positive, 5.0)
        self.assertEqual(result_negative, -5.0)

    def test_check_string_options_by_number_case_sensitive(self):
        """
        Tests that, passing a set of options, the correct value is returned if
        user enters the number of the item in the list (case sensitive scenario)
        """
        options = {'1': 'Oranges', '2': 'Apples', '3': 'Pineapples'}
        selected = 3
        expected_returned_value = options[str(selected)]

        result, valid = self.inputy.check_string(str(selected), options, '')

        self.assertTrue(valid)
        self.assertEqual(result, expected_returned_value)

    def test_check_string_options_by_number_case_insensitive(self):
        """
        Tests that, passing a set of options, the correct value is returned if
        user enters the number of the item in the list (case insensitive scenario)
        """
        options = {'1': 'Oranges', '2': 'Apples', '3': 'Pineapples'}
        selected = 3
        expected_returned_value = options[str(selected)]

        result, valid = self.inputy.check_string(str(selected), options, 'i')

        self.assertTrue(valid)
        self.assertEqual(result, expected_returned_value)

    def test_check_string_options_invalid_by_number_case_insensitive(self):
        """
        Tests that, passing a set of options, the correct value is returned if
        user enters the number of the item in the list (case insensitive scenario)
        """
        options = {'1': 'Oranges', '2': 'Apples', '3': 'Pineapples'}
        selected = 6  # invalid
        expected_returned_value = str(selected)

        result, valid = self.inputy.check_string(str(selected), options, 'i')

        self.assertFalse(valid)
        self.assertEqual(result, expected_returned_value)

    def test_check_string_options_case_insensitive(self):
        """
        Test that passing a different case for one of the options still
        returns the value
        """
        options = {'1': 'Oranges', '2': 'Apples', '3': 'Pineapples'}
        selected_capital_case = 'ORANGES'
        selected_lower_case = 'oranges'
        expected_returned_value = options['1']

        result_capital, valid_capital = self.inputy.check_string(selected_capital_case, options, 'i')
        result_lower, valid_lower = self.inputy.check_string(selected_lower_case, options, 'i')

        self.assertTrue(valid_capital)
        self.assertTrue(valid_lower)

        self.assertEqual(result_capital, expected_returned_value)
        self.assertEqual(result_lower, expected_returned_value)

    def test_check_string_case_sensitive(self):
        """
        Test that passing a different case for one of the options returns
        an invalid value
        """
        options = {'1': 'Oranges', '2': 'Apples', '3': 'Pineapples'}
        selected_capital_case = 'ORANGES'
        selected_matched_case = 'Oranges'
        expected_returned_value = options['1']

        result_capital, valid_capital = self.inputy.check_string(selected_capital_case, options, '')
        result_matched, valid_matched = self.inputy.check_string(selected_matched_case, options, '')

        self.assertFalse(valid_capital)
        self.assertTrue(valid_matched)

        self.assertEqual(result_matched, expected_returned_value)

    @mock.patch('builtins.input', return_value=str_valid_test)
    def test_passing_no_parameters_returns_a_value_str(self, mock_input):
        """ Testing 'inputy' as a normal 'input()' function """
        result_str = self.inputy.format_input()
        self.assertEqual(result_str, self.str_valid_test)

    @mock.patch('builtins.input', return_value=int_valid_test)
    def test_passing_no_parameters_returns_a_value_str_from_int(self, mock_input):
        """ Testing 'inputy' as a normal 'input()' function """
        result_str_from_int = self.inputy.format_input()
        self.assertEqual(result_str_from_int, str(self.int_valid_test))

    @mock.patch('builtins.input', side_effect=[str_valid_test, bool_valid_test, float_valid_test, None, int_valid_test])
    def test_passed_invalid_when_requested_int(self, mock_input):
        """
        Test that, when specifying the users has to enter an integer,
        the message is prompted until a valid number is passed
        """
        result_valid_int = self.inputy.format_input(type='int')

        self.assertEqual(result_valid_int, self.int_valid_test)

    @mock.patch('builtins.input', side_effect=[None, str_valid_test, bool_valid_test, float_valid_test])
    def test_passed_invalid_when_requested_float(self, mock_input):
        """
        Test that, when specifying the users has to enter a number,
        the message is prompted until a valid number is passed
        """
        result_valid_int = self.inputy.format_input(type='float')

        self.assertEqual(result_valid_int, self.float_valid_test)

    @mock.patch('builtins.input', side_effect=[str_valid_test, None, int_valid_test, bool_valid_test])
    def test_passed_invalid_when_requested_boolean(self, mock_input):
        """
        Test that, when specifying the user has to enter a boolean
        the message is prompted until a boolean is passed
        """
        result_valid_boolean = self.inputy.format_input(type='bool')

        self.assertEqual(result_valid_boolean, self.bool_valid_test)

    @mock.patch('builtins.input', side_effect=[str_valid_test, None, int_valid_test, 'True'])
    def test_passed_invalid_when_requested_boolean_str(self, mock_input):
        """
        Test that, when specifying the user has to enter a boolean
        the message is prompted until a test case insensitive with the name of
        one of the boolean values is passed
        """
        result_valid_boolean = self.inputy.format_input(type='bool')

        self.assertEqual(result_valid_boolean, True)

    @mock.patch('builtins.input', return_value=str_valid_test)
    def test_passing_and_invalid_input_type(self, mock_input):
        """
        Tests that passing and invalid input type raises an InvalidInputType
        exception. We mock the input() just in case the tests reaches that section
        """
        invalid_input_type = 'not_int_nor_float'
        with self.assertRaises(InvalidInputType) as e:
            self.inputy.format_input(type=invalid_input_type)
        self.assertEqual(e.exception.input_type, invalid_input_type)

    def test_converting_value_to_integer(self):
        number_as_string_value = '2'
        invalid_string = 'Not a number as a string'
        none_value = None

        self.assertEqual(Printy._to_int(number_as_string_value), 2)
        self.assertEqual(Printy._to_int(none_value), None)
        with self.assertRaises(ValueError):
            Printy._to_int(invalid_string)

    def test_max_digits(self):
        """ Tests restriction when passing a max_digits parameter """

        max_digits = 3
        invalid_number = '1234'
        valid_number = '123'

        invalid_number_value, invalid_number_valid_value = self.inputy._check_number(
            int, value=invalid_number, max_digits=max_digits
        )
        valid_number_value, valid_number_valid_value = self.inputy._check_number(
            int, valid_number, max_digits=max_digits
        )

        self.assertTrue(valid_number_valid_value)
        self.assertFalse(invalid_number_valid_value)

    def test_max_decimals(self):
        """ Tests restriction when passing a max_decimals parameter """

        max_decimals = 3
        invalid_number = '1234.1234'
        valid_number = '1234.123'

        invalid_number_value, invalid_number_valid_value = self.inputy._check_number(
            float, value=invalid_number, max_decimals=max_decimals
        )
        valid_number_value, valid_number_valid_value = self.inputy._check_number(
            float, valid_number, max_decimals=max_decimals
        )

        self.assertTrue(valid_number_valid_value)
        self.assertFalse(invalid_number_valid_value)


"""

Printy extends the capabilities of the builtins print() and input(), allowing
us to add some colors and formats to the text, either globally or inline with
an friendly and intuitive syntax, reading from a file and adding some validation
to the data the final user is entering by the console

"""

from .core import Printy
from .flags import Flags

__version__ = "2.1.1"

__all__ = ['raw_format', 'printy', 'inputy', 'COLORS', 'FORMATS']

printy_instance = Printy()

# If user just want to get the formatted text with the ANSI escape sequences
raw_format = printy_instance.get_formatted_text

# Main function to extend print() functionality
printy = printy_instance.format

# Main function to extend input() functionality
inputy = printy_instance.format_input

# shortcut to get a list of the available flags and formats
available_flags = Flags.get_flags().keys()
COLORS = list(filter(lambda c: c.islower(), available_flags))
FORMATS = list(filter(lambda f: f.isupper(), available_flags))



"""
renders the helper text that shows some of the available functions
by running python -m printy
"""

from . import printy

if __name__ == '__main__':
    from .helpme import helpme
    printy(helpme)




helpme = """
#########################################################
#########################################################

[bBU]PRINTY@

Printy lets you colorize and apply some standard formats to your text with
an intuitive and friendly API based on flags to specify the formats. You can
either apply a global format or inline formats to specific parts of your text!

[nB]Let's get started!@

[B]COLORS@

[gH]#### Gray Scale@
[k]'k' -> Applies a black color to the text@
[g]'g' -> Applies a grey color to the text@
[w]'w' -> Applies a white color to the text@

[rH]#### Red Scale@
[<r]'<r' -> Applies a darkred color to the text@
[r]'r'  -> Applies a red color to the text@
[r>]'r>' -> Applies a lightred color to the text@

[nH]#### Green Scale@
[<n]'<n' -> Applies a darkgreen color to the text@
[n]'n'  -> Applies a green color to the text@
[n>]'n>' -> Applies a lightgreen color to the text@

[yH]#### Yellow Scale@
[<y]'<y' -> Applies a darkyellow color to the text@
[y]'y'  -> Applies a yellow color to the text@
[y>]'y>' -> Applies a lightyellow color to the text@

[bH]#### Blue Scale@
[<b]'<b' -> Applies a darkblue color to the text@
[b]'b'  -> Applies a blue color to the text@
[b>]'b>' -> Applies a lightblue color to the text@

[mH]#### Magenta Scale@
[<m]'<m' -> Applies a darkmagenta color to the text@
[m]'m'  -> Applies a magenta color to the text@
[m>]'m>' -> Applies a lightmagenta color to the text@

[cH]#### Cyan Scale@
[<c]'<c' -> Applies a darkcyan color to the text@
[c]'c'  -> Applies a cyan color to the text@
[c>]'c>' -> Applies a lightcyan color to the text@

[oH]#### Orange Scale@
[<o]'<o' -> Applies a darkorange color to the text@
[o]'o'  -> Applies a orange color to the text@
[o>]'o>' -> Applies a lightorange color to the text@

[pH]#### Purple Scale@
[<p]'<p' -> Applies a darkpurple color to the text@
[p]'p'  -> Applies a purple color to the text@
[p>]'p>' -> Applies a lightpurple color to the text@

[B]FORMATS@

[B]'B' -> Applies a bold font weight to the text@
[U]'U' -> Applies an underline to the text@
[I]'I' -> Applies an italic font type to the text@
[H]'H' -> Highlights the text@
[S]'S' -> crosses out the text, aka Strike@
[D]'S' -> Dim effect@


[B]HOW TO USE IT?@

First import printy:

[n]    >>>@ [<o]from@ printy [<o]import@ printy

Now apply one or more of the available flags to the text like this:

[n]    >>>@ printy([n>]'Some text to be formatted'@[<o],@ [n>]'b'@)
    [b]Some text to be formatted@

[n]    >>>@ printy([n>]'Some text to be formatted'@[<o],@ [n>]'bHI'@)
    [bHI]Some text to be formatted@

Or as Inline format, use the \[\] to specify the flags, and the \@ to finish
the format for a specific section:

[n]    >>>@ printy([n>]'\[r\]Red\@ Default Color \[yH\]Yellow Highlighted\@ Default Color'@)
    [r]Red@ Default Color [yH]Yellow Highlighted@ Default Color

You can always override the whole format with a global flag:

[n]    >>>@ printy([n>]'\[r\]Red\@ Default Color \[yH\]Yellow Highlighted\@ Default Color'@[<o],@ [n>]'b'@)
    [b]Red Default Color Yellow Highlighted Default Color@

Or you can change only the predefined color:

[n]    >>>@ printy([n>]'\[r\]Red\@ Default Color \[yH\]Yellow Highlighted\@ Default Color'@[<o],@ [r]predefined@=[n>]'b'@)
    [r]Red@ [b]Default Color@ [yH]Yellow Highlighted@ [b]Default Color@

[n]    >>>@ printy([n>]'\[r\]Red\@ Default Color \[yH\]Yellow Highlighted\@ Default Color'@[<o],@ [r]predefined@=[n>]'nBIU'@)
    [r]Red@ [nBIU]Default Color@ [yH]Yellow Highlighted@ [nBIU]Default Color@

If you need to use one of the special characters ('\[', '\]', '\@'), simply escape them

[n]    >>>@ printy([n>]'\[B\]\\\[myemail\\\@mydomain.com\\\]\@'@)
    [B]\[myemail\@mydomain.com\]@

Now you can do stuffs like:

[n]    >>>@ text = [n>]'Hello world, python is awesome!!!'@
[n]    >>>@ printy(text.replace([n>]'python'@[<o],@ [n>]'\[r\]python\@'@))
    Hello world, [r]python@ is awesome

Or some html highlighting:

[n]    >>>@ html = (
[n]    ...@ [n>]'<div \[p\]class=\@\[n>\]"active"\@ \[p\]id=\@\[n>\]"my-div"\@>'@
[n]    ...@ [n>]'    <span \[p\]data-some-data=\@\[n>\]"extra-data"\@>\[p\]Some text\@</span>'@
[n]    ...@ [n>]'</div>'@)
[n]    >>>@ printy(html[<o],@ [r]default@=[n>]'y>'@)
    [y>]<div@ class=[n>]'active'@ id=[n>]'my-div'@[y>]>@
        [y>]<span@ data-some-data=[n>]'extra-data'@[y>]>@Some text[y>]</span>@
    [y>]</div>@

Or, you can use it with python's formatting strings as well:

[n]    >>>@ minutes = [c>]60@
[n]    >>>@ printy([n>]f'A day has \[y\]@[<o]{@minutes * [c>]24@[<o]}@[n>]\@ minutes'@)
    A day has [y]1400@ minutes

[B]What about @[p>B]input@()?

printy comes with a wrapper for the python built-in [p>]input@() function

[n]    >>>@ [<o]from@ printy [<o]import@ inputy

Try adding adding some formats the same way you did with printy and 
[y]Check this link for a full documentation!@

https://github.com/edraobdu/printy    
"""




#### Printy


class InvalidFlag(Exception):
    """ raised when an invalid flag is passed to the 'printy' object"""

    def __init__(self, flag):
        self.flag = flag

    def __str__(self):
        return "'%s' is not a valid flag" % self.flag


#### Inputy


class InvalidInputType(Exception):
    """ raised when an invalid 'type' is passed to the 'inputy' function"""

    def __init__(self, input_type):
        self.input_type = input_type

    def __str__(self):
        return "'%s' is not a valid type" % self.input_type
